#!/usr/bin/env python
# -*- coding: utf-8 -*-
import inspect
import logging
import numpy as np
import pandas as pd
from sklearn.metrics import accuracy_score, r2_score, mean_squared_error

from .base_accessor import BaseAccessor
from .features_utils import rmspe_score, neg_rmspe_score, neg_mae_score, neg_rmse_score
from .pre_learning_accessor import PreLearningAccessor


class BaselineClassifier(object):
	"""
	Classifier following the scikit-learn API and always predicting the most frequent class in-sample.
	"""
	def fit(self, x, y):
		values, counts = np.unique(y, return_counts=True)
		ind = np.argmax(counts)
		self.training_mode = values[ind]

	def predict(self, x):
		try:
			n = x.shape[0]
			return np.array([self.training_mode]*n)[:, None]

		except AttributeError:
			logging.error('The model should be fitted first')



class BaselineRegressor(object):
	"""
	Regressor following the scikit-learn API and always predicting the in-sample mean.
	"""
	def __init__(self, baseline='mean'):
		assert baseline in ['mean', 'median']
		self._baseline = np.nanmean if baseline == 'mean' else np.nanmedian


	def fit(self, x, y):
		self.training_baseline = self._baseline(y)

	def predict(self, x):
		try:
			n = x.shape[0]
			return np.ones((n, 1))*self.training_baseline

		except AttributeError:
			logging.error('The model should be fitted first')


@pd.api.extensions.register_dataframe_accessor("kxy_learning")
class LearningAccessor(BaseAccessor):
	"""
	Extension of the pandas.DataFrame class with various analytics for automatically training predictive models.

	This class defines the :code:`kxy_learning` `pandas accessor <https://pandas.pydata.org/pandas-docs/stable/development/extending.html>`_.

	All its methods defined are accessible from any DataFrame instance as :code:`df.kxy_learning.<method_name>`, so long as the :code:`kxy` python package is imported alongside :code:`pandas`. 
	"""

	def _update_target(self, target, predictions):
		'''
		'''
		if self.additive_learning:
			if self.problem_type == 'regression':
				return target-predictions if self.regression_error_type == 'additive' else target/predictions

			if self.problem_type == 'classification':
				return np.logical_not(target == predictions).astype(int)

		else:
			return target

	def _compound_predictions(self, previous_predictions, new_predictions):
		'''
		'''
		if self.additive_learning:
			if self.problem_type == 'regression':
				return previous_predictions+new_predictions if self.regression_error_type == 'additive' else \
					previous_predictions*new_predictions

			if self.problem_type == 'classification':
				return np.abs(previous_predictions-new_predictions)

		else:
			return new_predictions

	def fit(self, target_column, learner_func, problem_type=None, snr='auto', train_frac=0.8, random_state=0, \
			force_redo=False, max_n_features=None, min_n_features=None, start_n_features=1, anonymize=False, \
			benchmark_feature=None, missing_value_imputation=False, score='auto', n_down_perf_before_stop=1, \
			regression_baseline='mean', additive_learning=False, regression_error_type='additive'):
		"""
		Train a lean boosted supervised learner, bringing in variables one at a time, in decreasing order of importance (as per :code:`df.kxy.variable_selection`), until doing so no longer improves validation performance or another stopping criterion is met.

		Specifically, training proceeds as follows. First, KXY's model-free variable selection is run (i.e. :code:`df.kxy.variable_selection`). 

		Then we train a model (instance returned by :code:`learner_func`) using the :code:`start_n_features` most important feature/variable to predict the target (defined by :code:`target_column`).

		Then we consider using the second most important variable to fix the mistakes made by the previously trained model.
	
		If doing so improves performance on the validation set, we keep going until either performance no longer improves on the validation set :code:`n_down_perf_before_stop` consecutive times, or we've selected :code:`max_n_features` features.



		Parameters
		----------
		target_column : str
			The name of the column containing true labels.
		learner_func : function
			A function returning an instance of a base learner. They should define a :code:`fit(x, y)` method and a :code:`predict(x)` method.
		problem_type : None | 'classification' | 'regression'
			The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.
		snr : 'auto' | 'low' | 'high'
			Set to :code:`low` if the problem is difficult (i.e. has a low signal-to-noise ratio) or the number of rows is small relative to the number of columns. Only used for model-free variable selection.
		train_frac : float
			The fraction of rows used for training and validation.
		random_state : int
			The seed to use for random training/validation/testing split.
		force_redo : bool
			Fitted models are saved. Set this parameters to :code:`True` to ignore saved models and refit.
		min_n_features : int | None
			Boosting will not stop until at least this many features/explanatory variables are selected.
		max_n_features : int | None
			Boosting will stop as soon as this many features/explanatory variables are selected.
		start_n_features : int
			The number of most important features boosting will start with.
		anonymize : bool
			When set to true, your explanatory variables will never be shared with KXY (at no performance cost).
		benchmark_feature : str | None
			When not None, 'benchmark' performance metrics using this column as predictor will be reported in the output dictionary.
		missing_value_imputation : bool
			When set to True, replace missing values with medians.
		n_down_perf_before_stop : int
			Number of consecutive down performances to observe before boosting stops.
		regression_baseline : str (:code:`mean` | :code:`median`)
			Whether to use the unconditional mean or median as the best predictor in the absence of explanatory variables.
			Choosing the mean corresponds to minimizing the L2 norm, whereas choosing the median corresponds to minimizing the L1 norm.
		additive_learning : bool
			When a new variable is added, whether errors/residuals should be fixed or a new model should be learned from scratch.
		regression_error_type : str ('additive' | 'multiplicative')
			For regression problems with additive learning, this determines whether the final model should be additive (pruning tries to reduce regressor residuals) or multiplicative (i.e. pruning tries to bring the ratio between true and predicted labels as closed to 1 as possible).


		Returns
		-------
		result : dict
			Dictionary containing selected variables, as well as training, validation and testing performance.
		"""
		obj = self._obj
		assert inspect.isfunction(learner_func), 'learner_func should be a class'
		assert target_column in obj.columns, 'The target column should be a valid column'
		if problem_type is None:
			problem_type = 'classification' if self.is_discrete(target_column) else 'regression'
		assert problem_type in ('classification', 'regression')
		self.problem_type = problem_type
		self.additive_learning = additive_learning
		assert regression_error_type in ('additive', 'multiplicative')
		self.regression_error_type = regression_error_type

		for col in obj.columns:
			assert not self.is_categorical(col), 'All columns should be numeric'

		x_columns = [_ for _ in obj.columns if _ != target_column]
		if self.problem_type == 'classification':
			labels = set(list(obj[target_column].values.astype(int)))
			binary_labels = {0, 1}
			assert labels.issubset(binary_labels), 'Classification labels should either be 0 or 1'

		if benchmark_feature:
			assert benchmark_feature in obj.columns, 'The benchmark feature should be a valid column'
		self.benchmark_feature = benchmark_feature
		if score == 'auto':
			score = 'r2_score' if problem_type == 'regression' else 'accuracy_score'
		score_func = eval(score)

		if getattr(self, 'models', None) is None or force_redo:
			# 0. Train/Validation split
			self.target_column = target_column
			# Reserve 1-train_frac for testing, and train_frac for training and validation
			self.test_df = obj.sample(frac=1.-train_frac, random_state=random_state)
			self.train_val_df = obj.drop(self.test_df.index)
			# Reserve train_frac*train_frac for training [...]
			self.train_df = self.train_val_df.sample(frac=train_frac, random_state=random_state)
			# [...] and train_frac*(1-train_frac) for validation.
			self.val_df = self.train_val_df.drop(self.train_df.index)

			if missing_value_imputation:
				# Basic missing value imputation
				self.train_df.fillna(self.train_df.median(), inplace=True)
				self.val_df.fillna(self.train_df.median(), inplace=True)
				self.test_df.fillna(self.train_df.median(), inplace=True)

			# 1. Model-free variable selection
			vs_accessor = PreLearningAccessor(obj)
			self.variable_selection_results = vs_accessor.variable_selection(self.target_column, problem_type=self.problem_type, \
				snr=snr, anonymize=anonymize)
			self.variables = [_ for _ in self.variable_selection_results['Variable'].values if _.lower() != 'no variable']
			n_variables = len(self.variables)
			if max_n_features:
				n_variables = min(n_variables, max_n_features)

			# 2. Sequentially add variables in decreasing order of importance, each time with the aim to predict previous residuals.
			y_train_df = self.train_df[[self.target_column]]
			y_train = y_train_df.values.copy()
			target_train = y_train.copy()
			y_train_pred = None
			y_val_df = self.val_df[[self.target_column]]
			y_val = y_val_df.values.copy()
			target_val = y_val.copy()
			y_val_pred = None

			models = []
			max_var_ixs = []
			m = BaselineRegressor(baseline=regression_baseline) if problem_type == 'regression' else BaselineClassifier()
			x_train = self.train_df[self.variables[:1]].values.copy()
			x_val = self.val_df[self.variables[:1]].values.copy()
			m.fit(x_train, y_train)
			target_train_pred = m.predict(x_train)
			target_train_pred = target_train_pred if len(target_train_pred.shape) > 1 else target_train_pred[:, None]
			y_train_pred = target_train_pred.copy()
			target_val_pred = m.predict(x_val)
			target_val_pred = target_val_pred if len(target_val_pred.shape) > 1 else target_val_pred[:, None]
			y_val_pred = target_val_pred.copy()
			previous_score = score_func(y_val, target_val_pred)
			logging.info('Baseline score (%s): %.3f' % (score, previous_score))

			target_train = self._update_target(target_train, target_train_pred)
			target_val = self._update_target(target_val, target_val_pred)
			models = models+[m] if self.additive_learning else [m]
			max_var_ixs = max_var_ixs+[0] if self.additive_learning else [0]	

			self.start_n_features = start_n_features
			n_down_perf = 0
			for i in range(self.start_n_features, n_variables+1):
				vs = self.variables[:i]
				x_train = self.train_df[vs].values.copy()
				x_val = self.val_df[vs].values.copy()
				n_vars = x_train.shape[1] if len(x_train.shape) > 1 else 1

				# Create the new model
				m = learner_func(n_vars=n_vars)

				# Fit the new model
				m.fit(x_train, target_train)

				# Error predictions
				target_train_pred = m.predict(x_train)
				target_train_pred = target_train_pred if len(target_train_pred.shape) > 1 else target_train_pred[:, None]
				target_val_pred = m.predict(x_val)
				target_val_pred = target_val_pred if len(target_val_pred.shape) > 1 else target_val_pred[:, None]

				# Target predictions updates
				y_train_pred_ = self._compound_predictions(y_train_pred, target_train_pred)
				y_val_pred_ = self._compound_predictions(y_val_pred, target_val_pred)

				# New validation score
				val_score = score_func(y_val, y_val_pred_)
				if val_score > previous_score or (min_n_features and i<=min_n_features):
					y_train_pred = y_train_pred_.copy()
					y_val_pred = y_val_pred_.copy()
					n_down_perf = 0
					logging.info('Variable #%d (%s) increased validation performance from %.3f to %.3f' % (i, self.variables[i-1], previous_score, val_score))
					previous_score = val_score

					target_train = self._update_target(target_train, target_train_pred)
					target_val = self._update_target(target_val, target_val_pred)
					models = models+[m] if self.additive_learning else [m]
					max_var_ixs = max_var_ixs+[i] if self.additive_learning else [i]					

				else:
					n_down_perf += 1
					logging.info('Validation performance did not increase for the %d-th consecutive time. Old: %.3f, New: %.3f, Variable: %s' % (n_down_perf, previous_score, val_score, self.variables[i-1]))
					if n_down_perf >= n_down_perf_before_stop:
						# Only stop after a certain number of consecutive down performance
						logging.info('Stopping training as validation performance did not increase %d consecutive times.' % n_down_perf_before_stop)
						break

				if max_n_features and (i==max_n_features):
					logging.info('Stopping training as the maximum number of variables (%d) has been reached' % max_n_features)
					break				

			self.models = models
			self.max_var_ixs = max_var_ixs
			self.selected_variables = self.variables[:self.max_var_ixs[-1]]

			# Compute training/validation/testing performances
			x_train_df = self.train_df[x_columns]
			y_train_df = self.train_df[[target_column]]
			train_predictions = self.predict(x_train_df)
			self.train_score = score_func(y_train_df.values, train_predictions.values)
			if self.problem_type == 'regression':
				self.train_rmse = mean_squared_error(y_train_df.values.flatten(), train_predictions.values.flatten(), squared=False)
				self.train_rmspe = rmspe_score(y_train_df.values.flatten(), train_predictions.values.flatten())
				self.train_r2 = r2_score(y_train_df.values.flatten(), train_predictions.values.flatten())

				if self.benchmark_feature:
					train_benchmark = self.train_df[self.benchmark_feature]
					self.train_benchmark_score = score_func(y_train_df.values, train_benchmark.values)
					self.train_benchmark_rmse = mean_squared_error(y_train_df.values, train_benchmark.values, squared=False)
					self.train_benchmark_rmspe = rmspe_score(y_train_df.values, train_benchmark.values)
					self.train_benchmark_r2 = r2_score(y_train_df.values.flatten(), train_benchmark.values.flatten())

			x_val_df = self.val_df[x_columns]
			y_val_df = self.val_df[[target_column]]
			val_predictions = self.predict(x_val_df)
			self.val_score = score_func(y_train_df.values, train_predictions.values)
			if self.problem_type == 'regression':
				self.val_rmse = mean_squared_error(y_val_df.values.flatten(), val_predictions.values.flatten(), squared=False)
				self.val_rmspe = rmspe_score(y_val_df.values.flatten(), val_predictions.values.flatten())
				self.val_r2 = r2_score(y_val_df.values.flatten(), val_predictions.values.flatten())

				if self.benchmark_feature:
					val_benchmark = self.val_df[self.benchmark_feature]
					self.val_benchmark_score = score_func(y_val_df.values, val_benchmark.values)
					self.val_benchmark_rmse = mean_squared_error(y_val_df.values, val_benchmark.values, squared=False)
					self.val_benchmark_rmspe = rmspe_score(y_val_df.values, val_benchmark.values)
					self.val_benchmark_r2 = r2_score(y_val_df.values.flatten(), val_benchmark.values.flatten())
				
			x_test_df = self.test_df[x_columns]
			y_test_df = self.test_df[[target_column]]
			test_predictions = self.predict(x_test_df)
			self.test_score = score_func(y_test_df.values, test_predictions.values)
			if self.problem_type == 'regression':
				self.test_rmse = mean_squared_error(y_test_df.values.flatten(), test_predictions.values.flatten(), squared=False)
				self.test_rmspe = rmspe_score(y_test_df.values.flatten(), test_predictions.values.flatten())
				self.test_r2 = r2_score(y_test_df.values.flatten(), test_predictions.values.flatten())

				if self.benchmark_feature:
					test_benchmark = self.test_df[self.benchmark_feature]
					self.test_benchmark_score = score_func(y_test_df.values, test_benchmark.values)
					self.test_benchmark_rmse = mean_squared_error(y_test_df.values, test_benchmark.values, squared=False)
					self.test_benchmark_rmspe = rmspe_score(y_test_df.values, test_benchmark.values)
					self.test_benchmark_r2 = r2_score(y_test_df.values.flatten(), test_benchmark.values.flatten())


		results = {'Selected Variables': self.selected_variables}
		if self.problem_type == 'regression':
			results['Training R-Squared'] = '%.3f' % self.train_r2
			results['Validation R-Squared'] = '%.3f' % self.val_r2
			results['Testing R-Squared'] = '%.3f' % self.test_r2
			results['Training RMSE'] = '%.5f' % self.train_rmse
			results['Validation RMSE'] = '%.5f' % self.val_rmse
			results['Testing RMSE'] = '%.5f' % self.test_rmse
			results['Training RMSPE'] = '%.5f' % self.train_rmspe
			results['Validation RMSPE'] = '%.5f' % self.val_rmspe
			results['Testing RMSPE'] = '%.5f' % self.test_rmspe
			if self.benchmark_feature:
				results['Benchmark Training R-Squared'] = '%.3f' % self.train_benchmark_r2
				results['Benchmark Validation R-Squared'] = '%.3f' % self.val_benchmark_r2
				results['Benchmark Testing R-Squared'] = '%.3f' % self.test_benchmark_r2
				results['Benchmark Training RMSE'] = '%.5f' % self.train_benchmark_rmse
				results['Benchmark Validation RMSE'] = '%.5f' % self.val_benchmark_rmse
				results['Benchmark Testing RMSE'] = '%.5f' % self.test_benchmark_rmse
				results['Benchmark Training RMSPE'] = '%.5f' % self.train_benchmark_rmspe
				results['Benchmark Validation RMSPE'] = '%.5f' % self.val_benchmark_rmspe
				results['Benchmark Testing RMSPE'] = '%.5f' % self.test_benchmark_rmspe

		if self.problem_type == 'classification':
			results['Training Accuracy'] = '%.3f' % self.train_score
			results['Validation Accuracy'] = '%.3f' % self.val_score
			results['Testing Accuracy'] = '%.3f' % self.test_score

		return results
		


	def predict(self, obj):
		"""
		Make predictions using the fitted model.


		Parameters
		----------
		obj : pandas.DataFrame
			A dataframe containing test explanatory variables about which we want to make predictions.


		Returns
		-------
		result : pandas.DataFrame
			A dataframe with the same index as :code:`obj`, and with one column whose name is the :code:`target_column` used for training.
		"""
		assert hasattr(self, 'models'), 'The model should be first fitted'
		for col in obj.columns:
			assert not obj.kxy.is_categorical(col), 'All columns should be numeric'

		data = obj[self.variables]

		if self.additive_learning:
			x = data[self.variables[:1]].values.copy()
			y_error_pred = self.models[0].predict(x)
			y_error_pred = y_error_pred if len(y_error_pred.shape) > 1 else y_error_pred[:, None]
			y_pred = y_error_pred

			for i in range(1, len(self.models)):
				vs = self.variables[:self.max_var_ixs[i]]
				x = data[vs].values.copy()
				y_error_pred = self.models[i].predict(x)
				y_error_pred = y_error_pred if len(y_error_pred.shape) > 1 else y_error_pred[:, None]
				y_pred = self._compound_predictions(y_pred, y_error_pred)

		else:
			x = data[self.selected_variables].values.copy()
			y_pred = self.models[0].predict(x)
			y_pred = y_pred if len(y_pred.shape) > 1 else y_pred[:, None]

		predictions = pd.DataFrame(y_pred, columns=[self.target_column], index=data.index)

		return predictions








#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
==================
kxy.api.decorators
==================
"""

from functools import wraps
import json
import os

def get_api_key():
	"""
	Retrieves the store API key, or None if none was provided.
	"""
	home = os.path.expanduser("~")
	path = os.path.join(home, '.kxy')
	file_name = os.path.join(path, 'config')
	try:
		with open(file_name, 'r') as f:
			config = json.load(f)
			existing_key = config.get('KXY_API_KEY', None)
			return existing_key
	except:
		return os.environ.get('KXY_API_KEY', None)

	return None



def has_api_key():
	"""
	Returns whether or not an API key was provided as a result of running :code:`kxy configure`.
	"""
	return get_api_key() is not None



def requires_api_key(method):
	"""
	Decorator used to make functions and methods calls fail
	when they require an API key and the user did not provide on 
	by running :code:`kxy configure`. The decorated function or method 
	is otherwise not affected.

	Raises
	------
	AssertionError
		If an API key was not previously recorded.
	"""
	@wraps(method)
	def wrapper(*args, **kw):		
		assert has_api_key(), "An API key should be provided. Please run 'kxy configure'"
		return method(*args, **kw)

	return wrapper


#!/usr/bin/env python
# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from pandarallel import pandarallel

from .base_accessor import BaseAccessor
from .features_utils import nanskew, nankurtosis, q25, q75, mode, modefreq, nextmode, nextmodefreq, lastmode, lastmodefreq, \
	nanmin, nanmax, nanmean, nanstd, nanmedian, nanmaxmmin, nansum, nanskewabs, nankurtosisabs, q25abs, q75abs,  \
	nanminabs, nanmaxabs, nanmeanabs, nanstdabs, nanmedianabs, nanmaxmminabs, nansumabs, n_unique


@pd.api.extensions.register_dataframe_accessor("kxy_features")
class FeaturesAccessor(BaseAccessor):
	"""
	Extension of the pandas.DataFrame class with various feature engineering functionalities.

	This class defines the :code:`kxy_features` `pandas accessor <https://pandas.pydata.org/pandas-docs/stable/development/extending.html>`_.

	All its methods defined are accessible from any DataFrame instance as :code:`df.kxy_features.<method_name>`, so long as the :code:`kxy` python package is imported alongside :code:`pandas`. 
	"""
	def ordinally_encode(self, target_column=None, method='one_hot'):
		"""
		Encode categorical (non-numeric) data.

		Parameters
		----------
		target_column : str
			The name of the column containing labels. When this column is categorical, each label is replaced by a distinct integer.
		method : 'one_hot' (default) | 'binary'
			Whether to use one-hot encoding or binary encoding to encode categorical variables.


		Returns
		-------
		result : pandas.DataFrame
			The ordinarily encoded dataframe.
		"""
		assert target_column is None or target_column in self._obj.columns, 'The target_column should be a column'
		assert method.lower() in ['one_hot', 'binary'], 'The encoding method should be either one_hot or binary'

		cat_columns = [col for col in self._obj.columns if col is not target_column and self.is_categorical(col)]
		
		if cat_columns:
			num_columns = [col for col in self._obj.columns if col is not target_column and col not in cat_columns]
			if method.lower() == 'one_hot':
				# One-hot encode categorical explanatory variables
				cat_x_encoded = pd.get_dummies(self._obj[cat_columns], prefix=cat_columns)

			if method.lower() == 'binary':
				# Binary encode categorical explanatory variables
				cat_x_encoded = pd.concat([self._binary_encode(col) for col in cat_columns], axis=1)

			if target_column:
				df_encoded = pd.concat([self._obj[[target_column]+num_columns], cat_x_encoded], axis=1)
			else:
				df_encoded = pd.concat([self._obj[num_columns], cat_x_encoded], axis=1)

			# Ordinarily encode the target if needed (e.g. for classification with non-numeric classes)
			if target_column and self.is_categorical(target_column):
				all_classes = sorted(list(set(list(self._obj[target_column]))))
				classes_map = {all_classes[i]: i for i in range(len(all_classes))}
				df_encoded[target_column] = df_encoded[target_column].apply(lambda c: classes_map[c])

		else:
			df_encoded = self._obj

		return df_encoded.astype(float)


	def _binary_encode(self, column):
		"""
		Binary-encode a specific column.
		"""
		assert self.is_categorical(column), 'The column should be categorical'
		x = self._obj[column].values
		x_ = np.array([str(_) for _ in x]).astype(str)
		n = len(x_)
		cats = set(x_)
		cats = list(sorted(cats))
		q = len(cats)
		h = int(np.ceil(np.log2(q)))
		cats_map = {cats[i]: i for i in range(q)}
		res = np.array([[int(_) for _ in list(bin(cats_map[x_[i]])[2:].zfill(h))] for i in range(n)])

		if len(res.shape) == 1:
			res = res[:, None]

		columns = ['%s_bit_%d' % (column, i) for i in range(h)]
		res = pd.DataFrame(res, columns=columns)

		return res


	def entity_features(self, entity, exclude=[], entity_name='*', filter_target=None, filter_target_gt=None, filter_target_lt=None, \
			include_filter_target=False):
		"""
		Group rows corresponding to the same entity and apply aggregation functions.

		For each ordinal column, we apply the following aggregation functions to rows corresponding to the same entity: mean, standard deviation, median, skewness, kurtosis, 25th and 75th percentiles, minimum, maximum, and the difference betwween maximum and minimum.

		For each non-ordinal column, we apply the following aggregation functions to rows corresponding to the same entity: mode and its frequency, second most frequent label and its frequency, least frequent label and its frequency.


		Parameters
		----------
		entity : str
			The column mapping rows to entities.
		exclude : list
			A list of columns to exclude from feature transformations.
		filter_target : str | None
			When specified, this is a column based on which we need to restrict the dataframe before generating features.
		filter_target_gt : str | None
			When specified, only rows with :code:`filter_target` greater than :code:`filter_target_gt` will be considered for feature generation.
		filter_target_lt : str | None
			When specified, only rows with :code:`filter_target` smaller than :code:`filter_target_gt` will be considered for feature generation.
		include_filter_target : bool
			Whether to use :code:`filter_target` for features generation.


		Returns
		-------
		result : pandas.DataFrame
			The dataframe of features.
		"""
		assert entity in self._obj.columns, 'The entity %s should be valid column' % entity
		_columns = self._obj.columns
		if filter_target and not include_filter_target:
			_columns = [_ for _ in _columns if _ != filter_target]

		cat_columns = list(set([col for col in _columns if self.is_categorical(col) and col != entity and col not in exclude]))
		ord_columns = list(set([col for col in _columns if not self.is_categorical(col) and col != entity and col not in exclude]))
		if ord_columns:
			mix_sgn_columns = [col for col in ord_columns if self._obj[col].lt(0).any() and self._obj[col].gt(0).any()]

		columns = cat_columns + ord_columns

		dfs = []
		agg = {}
		# Aggregation of categorical variables
		if cat_columns:
			# Most frequent value and it's frequency of occurence
			agg.update({'%s__%s.MODE()' % (entity, col): (col, mode) for col in cat_columns})
			agg.update({'%s__%s.MODEFREQ()' % (entity, col): (col, modefreq) for col in cat_columns})

			# Second most frequent value and it's frequency of occurence
			agg.update({'%s__%s.NEXTMODE()' % (entity, col): (col, nextmode) for col in cat_columns})
			agg.update({'%s__%s.NEXTMODEFREQ()' % (entity, col): (col, nextmodefreq) for col in cat_columns})

			# Least frequent value and it's frequency of occurence
			agg.update({'%s__%s.LASTMODE()' % (entity, col): (col, lastmode) for col in cat_columns})
			agg.update({'%s__%s.LASTMODEFREQ()' % (entity, col): (col, lastmodefreq) for col in cat_columns})

			# Number of unique values
			agg.update({'%s__%s.NUM_UNIQUE()' % (entity, col): (col, n_unique) for col in cat_columns})


		# Aggregation of ordinal variables
		if ord_columns:
			agg.update({'%s__%s.SUM()' % (entity, col): (col, nansum) for col in ord_columns})
			agg.update({'%s__%s.MEAN()' % (entity, col): (col, nanmean) for col in ord_columns})
			agg.update({'%s__%s.STD()' % (entity, col): (col, nanstd) for col in ord_columns})
			agg.update({'%s__%s.MEDIAN()' % (entity, col): (col, nanmedian) for col in ord_columns})
			agg.update({'%s__%s.SKEW()' % (entity, col): (col, nanskew) for col in ord_columns})
			agg.update({'%s__%s.KURT()' % (entity, col): (col, nankurtosis) for col in ord_columns})
			agg.update({'%s__%s.Q25()' % (entity, col): (col, q25) for col in ord_columns})
			agg.update({'%s__%s.Q75()' % (entity, col): (col, q75) for col in ord_columns})
			agg.update({'%s__%s.MIN()' % (entity, col): (col, nanmin) for col in ord_columns})
			agg.update({'%s__%s.MAX()' % (entity, col): (col, nanmax) for col in ord_columns})
			agg.update({'%s__%s.MAX-MIN()' % (entity, col): (col, nanmaxmmin) for col in ord_columns})

			if mix_sgn_columns:
				agg.update({'%s__%s.ABS().SUM()' % (entity, col): (col, nansumabs) for col in mix_sgn_columns})
				agg.update({'%s__%s.ABS().MEAN()' % (entity, col): (col, nanmeanabs) for col in mix_sgn_columns})
				agg.update({'%s__%s.ABS().STD()' % (entity, col): (col, nanstdabs) for col in mix_sgn_columns})
				agg.update({'%s__%s.ABS().MEDIAN()' % (entity, col): (col, nanmedianabs) for col in mix_sgn_columns})
				agg.update({'%s__%s.ABS().SKEW()' % (entity, col): (col, nanskewabs) for col in mix_sgn_columns})
				agg.update({'%s__%s.ABS().KURT()' % (entity, col): (col, nankurtosisabs) for col in mix_sgn_columns})
				agg.update({'%s__%s.ABS().Q25()' % (entity, col): (col, q25abs) for col in mix_sgn_columns})
				agg.update({'%s__%s.ABS().Q75()' % (entity, col): (col, q75abs) for col in mix_sgn_columns})
				agg.update({'%s__%s.ABS().MIN()' % (entity, col): (col, nanminabs) for col in mix_sgn_columns})
				agg.update({'%s__%s.ABS().MAX()' % (entity, col): (col, nanmaxabs) for col in mix_sgn_columns})
				agg.update({'%s__%s.ABS().MAX-MIN()' % (entity, col): (col, nanmaxmminabs) for col in mix_sgn_columns})
		
		# Number of rows per entity
		agg.update({'COUNT(%s)' % entity_name: (columns[0], 'count')})

		# Filter if necessary
		obj = self._obj.copy()
		if filter_target:
			assert filter_target in obj.columns, 'The filter column should be a valid column'
			if filter_target_gt:
				obj = obj[obj[filter_target] > filter_target_gt]

			if filter_target_lt:
				obj = obj[obj[filter_target] < filter_target_lt]

		# Features
		entity_grp = obj.groupby(entity)
		df = entity_grp.agg(**agg)

		return df


	def deviation_features(self, exclude=[], means=None, quantiles=None, return_baselines=False):
		"""
		Extend the dataframe with deviations of ordinal columns from row-wise aggregtes such as mean, median, 25th and 75th percentiles.


		Parameters
		----------
		exclude : list
			A list of columns to exclude from feature transformations.
		means : pandas.DataFrame | None
			Which values, if any, to use as means.
		quantiles : pandas.DataFrame | None
			Which values, if any, to use as 25th, 50th, 75th percentiles.
		return_baselines : bool
			Whether to return which baselines have been used.


		Returns
		-------
		result : pandas.DataFrame
			The original dataframe extended with computed features.
		"""
		ord_columns = [col for col in self._obj.columns if not self.is_categorical(col) and col not in exclude]

		if means is None:
			means = self._obj.mean(axis=0, skipna=True)

		if quantiles is None:
			quantiles = self._obj.quantile(q=[0.25, 0.5, 0.75])

		df = self._obj.copy()
		if ord_columns:
			for col in ord_columns:
				df['%s.ABS(* - MEAN(*))' % col] = np.abs(df[col]-means.loc[col])
				df['%s.ABS(* - MEDIAN(*))' % col] = np.abs(df[col]-quantiles.loc[0.5][col])
				df['%s.ABS(* - Q25(*))' % col] = np.abs(df[col]-quantiles.loc[0.25][col])
				df['%s.ABS(* - Q75(*))' % col] = np.abs(df[col]-quantiles.loc[0.75][col])

		if return_baselines:
			return df, means, quantiles
		else:
			return df


	def temporal_features(self, max_lag=10, exclude=[], index=None, groupby=None, sort_by=None):
		"""
		Extend the dataframe with some rolling statistics (e.g. rolling average, rolling min, rolling max, rolling max-rolling min, etc.) for all lags from 2 to the configured maximum lag.


		Parameters
		----------
		exclude : list
			A list of columns to exclude from feature transformations.
		max_lag : int
			The largest lag to consider.
		index : str | None (default)
			The column, if any, to set as index and sort before computing rolling statistics.
		groupby : str | None
			If provided, we will use this column to perform a groupby before temporal aggregation.
		sort_by : str | list | None
			Columns, if any, we need to sort the dataframe by, prior to rolling.


		Returns
		-------
		result : pandas.DataFrame
			The original dataframe extended with computed temporal features.
		"""
		if index:
			exclude += [index]

		if groupby:
			exclude += [groupby]

		ord_columns = [col for col in self._obj.columns if not self.is_categorical(col) and col not in exclude]
		df = self._obj.copy()
		if index and (df.index.name != index):
			df = df.set_index(index)

		if sort_by:
			df = df.sort_values(by=sort_by)

		dfs = [df.copy()]

		if groupby:
			pandarallel.initialize()
			def apply_func(s):
				''' '''
				lag = max_lag+1
				col_map = {}
				col_map.update({'%s_nanmean' % col: '%s.GROUPBY(%s).LAST(%d).MEAN()' % (col, groupby, lag) for col in ord_columns})
				col_map.update({'%s_nanmin' % col: '%s.GROUPBY(%s).LAST(%d).MIN()' % (col, groupby, lag) for col in ord_columns})
				col_map.update({'%s_nanmax' % col: '%s.GROUPBY(%s).LAST(%d).MAX()' % (col, groupby, lag) for col in ord_columns})
				col_map.update({'%s_nanmaxmmin' % col: '%s.GROUPBY(%s).LAST(%d).MAX-MIN()' % (col, groupby, lag) for col in ord_columns})
				cols = [_ for _ in col_map.values()]
				res = s.rolling(lag, min_periods=1).aggregate([nanmean, nanmin, nanmax, nanmaxmmin])
				res.columns = res.columns.map('_'.join).to_series().map(col_map)
				return res[cols]

			feat_df = df.groupby(groupby, sort=False).parallel_apply(apply_func)
			feat_df.reset_index(inplace=True)
			if df.index.name:
				feat_df.set_index(df.index.name, inplace=True)
			else:
				feat_df.set_index('index', inplace=True)
				feat_df.index.name = df.index.name
			dfs += [feat_df]
			df = pd.concat(dfs, axis=1)

		else:
			lag = max_lag+1
			col_map = {}
			col_map.update({'%s_nanmean' % col: '%s.LAST(%d).MEAN()' % (col, lag) for col in ord_columns})
			col_map.update({'%s_nanmin' % col: '%s.LAST(%d).MIN()' % (col, lag) for col in ord_columns})
			col_map.update({'%s_nanmax' % col: '%s.LAST(%d).MAX()' % (col, lag) for col in ord_columns})
			col_map.update({'%s_nanmaxmmin' % col: '%s.LAST(%d).MAX-MIN()' % (col, lag) for col in ord_columns})
			cols = [_ for _ in col_map.values()]
			feat_df = df.rolling(lag, min_periods=1).aggregate([nanmean, nanmin, nanmax, nanmaxmmin])
			feat_df.columns = feat_df.columns.map('_'.join).to_series().map(col_map)
			dfs += [feat_df[cols]]
			df = pd.concat(dfs, axis=1)

		return df


	def process_time_columns(self, columns):
		"""
		Extract features from timestamp columns such as: Month, Day, Day of Week, Hour, AM/PM.


		Parameters
		----------
		columns : list
			The list of columns that should be interprated as UTC epoch timestamps.


		Returns
		-------
		result : pandas.DataFrame
			The features dataframe (does not include the original dataframe)
		"""
		res = pd.DataFrame(index=self._obj.index)
		for col in columns:
			times = pd.to_datetime(self._obj[col], unit="s", infer_datetime_format=True)
			res['%s.HOUR()' % col] = times.dt.hour
			res['%s.DAYOFWEEK()' % col] = times.dt.dayofweek
			res['%s.DAY()' % col] = times.dt.day
			res['%s.MONTH()' % col] = times.dt.month
			res['%s.YEAR()' % col] = times.dt.year

		return res


	def generate_features(self, entity=None, encoding_method='one_hot', index=None, max_lag=None, exclude=[], \
			means=None, quantiles=None, return_baselines=False, entity_name='*', filter_target=None, \
			filter_target_gt=None, filter_target_lt=None, include_filter_target=False, fill_na=False, \
			temporal_groupby=None, temporal_sort_by=None, time_columns=None):
		"""
		Generate a wide range of candidate features to search from.

		We first compute entity features if needed. 

		Then we extend the resulting dataframe with deviations of ordinal columns from row-wise aggregtes such as mean, median, 25th and 75th percentiles. 

		Finally, we ordinally-encode the resulting dataframe and apply temporal transformations if required.


		Parameters
		----------
		entity : str
			The column mapping rows to entities.
		filter_target : str | None
			When specified, this is a column based on which we need to restrict the dataframe before generating entity features.
		filter_target_gt : str | None
			When specified, only rows with :code:`filter_target` greater than :code:`filter_target_gt` will be considered for entity feature generation.
		filter_target_lt : str | None
			When specified, only rows with :code:`filter_target` smaller than :code:`filter_target_gt` will be considered for entity feature generation.
		include_filter_target : bool
			Whether to use :code:`filter_target` for features generation.
		encoding_method : 'one_hot' (default) | 'binary'
			The encoding method to use for categorical variables.
		exclude : list
			A list of columns to exclude from feature transformations.
		max_lag : int | None
			The largest lag, if any, to consider for temporal features. Set to None to avoid temporal features.
		index : str | None (default)
			The column, if any, to set as index and sort before computing temporal features.
		means : pandas.DataFrame | None
			Which values, if any, to use as means for deviation features.
		quantiles : pandas.DataFrame | None
			Which values, if any, to use as 25th, 50th, 75th percentiles for deviation features.
		return_baselines : bool
			Whether to return which baselines have been used for deviation features.
		temporal_groupby : str | None
			If provided, we will use this column to perform a groupby before temporal aggregation.
		temporal_sort_by : str | None
			If provided, we will use this column to sort the dataframe before rolling when computing temporal features.
		time_columns : list | None
			The list of columns that correspond to times and from which we should extract features such as hour, day of week etc.


		Returns
		-------
		result : pandas.DataFrame
			The original dataframe extended with computed temporal features.
		"""
		accessor = self
		if entity:
			# Entity features
			df = accessor.entity_features(entity, entity_name=entity_name, filter_target=filter_target, \
				filter_target_gt=filter_target_gt, filter_target_lt=filter_target_lt, \
				include_filter_target=include_filter_target)
			accessor = FeaturesAccessor(df)

		if index:
			exclude += [index]
			
		if temporal_groupby:
			exclude += [temporal_groupby]

		if time_columns:
			# Extract information from time columns such as hour, day of week etc.
			df = accessor.process_time_columns(time_columns)
			accessor = FeaturesAccessor(df)

		# Deviation features
		res = accessor.deviation_features(exclude=exclude, means=means, quantiles=quantiles, return_baselines=return_baselines)
		df = res[0] if return_baselines else res
		accessor = FeaturesAccessor(df)

		# Ordinally encode
		df = accessor.ordinally_encode(method=encoding_method)
		accessor = FeaturesAccessor(df)

		if max_lag:
			# Temporal/trend features
			df = accessor.temporal_features(exclude=exclude, max_lag=max_lag, index=index, groupby=temporal_groupby, sort_by=temporal_sort_by)

		if fill_na:
			df = df.fillna(df.median(skipna=True))

		if return_baselines:
			return df, res[1], res[2]
		else:
			return df






#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
logging.basicConfig(level=logging.INFO)
import requests
import sys
from time import time, sleep

import numpy as np
import pandas as pd

try:
	get_ipython().__class__.__name__
	from halo import HaloNotebook as Halo
except:
	from halo import Halo

from kxy.api import APIClient, upload_data

# Cache old job ids to avoid being charged twice for the same job.
IACORR_JOB_IDS = {}

def information_adjusted_correlation(data_df, market_column, asset_column):
	"""
	Estimate the information-adjusted correlation between an asset return :math:`r` and the market return :math:`r_m`: :math:`\\text{IA-Corr}\\left(r, r_m \\right) := \\text{sgn}\\left(\\text{Corr}\\left(r, r_m \\right) \\right) \\left[1 - e^{-2I(r, r_m)} \\right]`, where :math:`\\text{sgn}\\left(\\text{Corr}\\left(r, r_m \\right) \\right)` the sign of the Pearson correlation coefficient.

	Unlike Pearson's correlation coefficient, which is 0 if and only if asset return and market return are **decorrelated** (i.e. they exhibit no linear relation), information-adjusted correlation is 0 if and only if market and asset returns are **statistically independent** (i.e. the exhibit no relation, linear or nonlinear).


	Parameters
	----------
	data_df : pandas.DataFrame
		The pandas DataFrame containing the data.
	market_column : str
		The name of the column containing market returns.
	asset_column : str
		The name of the column containing asset returns.


	Returns
	-------
	result : float
		The information-adjusted correlation.

	"""
	assert market_column in data_df.columns, 'The market column should be a column of the dataframe.'
	assert asset_column in data_df.columns, 'The asset column should be a column of the dataframe.'
	assert np.can_cast(data_df[market_column], float), 'The market return column should be numeric'
	assert np.can_cast(data_df[asset_column], float), 'The asset return column should be numeric'

	k = 0
	kp = 0
	max_k = 100
	spinner = Halo(text='Waiting for results from the backend.', spinner='dots')
	spinner.start()

	df = data_df[[market_column, asset_column]]
	file_name = upload_data(df)
	if file_name:
		job_id = IACORR_JOB_IDS.get(file_name, None)

		if job_id:
			api_response = APIClient.route(
				path='/wk/ia-corr', method='POST', 
				file_name=file_name, market_column=market_column, \
				asset_column=asset_column, \
				timestamp=int(time()), job_id=job_id)
		else:
			api_response = APIClient.route(
				path='/wk/ia-corr', method='POST', \
				file_name=file_name, market_column=market_column, \
				asset_column=asset_column, \
				timestamp=int(time()))

		initial_time = time()
		while api_response.status_code == requests.codes.ok and k < max_k:
			if kp%2 != 0:
				sleep(2 if kp<5 else 5 if k < max_k-4 else 300)
				kp += 4
				k = kp//2
			else:
				try:
					response = api_response.json()
					if 'job_id' in response:
						job_id = response['job_id']
						IACORR_JOB_IDS[file_name] = job_id
						sleep(2 if kp<5 else 5 if k < max_k-4 else 300)
						kp += 4
						k = kp//2

						# Note: it is important to pass the job_id to avoid being charged twice for the same work.
						api_response = APIClient.route(
							path='/wk/ia-corr', method='POST', 
							file_name=file_name, market_column=market_column, \
							asset_column=asset_column, \
							timestamp=int(time()), job_id=job_id)

						try:
							response = api_response.json()
							if 'eta' in response:
								progress_text = '%s%% Completed.' % response['progress_pct'] if 'progress_pct' in response else ''
								spinner.text = 'Waiting for results from the backend. ETA: %s. %s' % (response['eta'], progress_text)
						except:
							pass

					if 'job_id' not in response:
						duration = int(time()-initial_time)
						duration = str(duration) + 's' if duration < 60 else str(duration//60) + 'min'
						spinner.text = 'Received results from the backend in %s' % duration
						spinner.succeed()

						if 'ia-corr' in response:
							return response['ia-corr']
						else:
							return np.nan

				except:
					spinner.text = 'The backend encountered an unexpected error we are looking into. Please try again later.'
					spinner.fail()
					logging.exception('\nInformation-adjusted correlation failed. Last HTTP code: %s' % api_response.status_code)
					return None


		if api_response.status_code != requests.codes.ok:
			spinner.text = 'The backend is taking longer than expected. Please try again later.'
			spinner.fail()
			try:
				response = api_response.json()
				if 'message' in response:
					logging.error('\n%s' % response['message'])
			except:
				logging.error('\nInformation-adjusted correlation failed. Last HTTP code: %s' % api_response.status_code)

	return None






#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Estimation of the top-:math:`k` most valuable variables in a supervised learning problem for every possible :math:`k`, and 
the corresponding achievable performances.
"""
import logging
logging.basicConfig(level=logging.INFO)
import requests
import sys
from time import time, sleep

import numpy as np
import pandas as pd

try:
	get_ipython().__class__.__name__
	from halo import HaloNotebook as Halo
except:
	from halo import Halo

from kxy.api import APIClient, upload_data

# Cache old job ids to avoid being charged twice for the same job.
EXPLANATION_JOB_IDS = {}

def model_explanation(data_df, prediction_column, problem_type, snr='auto'):
	"""
	.. _model-explanation:
	Analyzes the variables that a model relies on the most in a brute-force fashion.
	
	The first variable is the variable the model relies on the most. The second variable is the variable that complements the first variable the most in explaining model decisions etc.

	Running performances should be understood as the performance achievable when trying to guess model predictions using variables with selection order smaller or equal to that of the row.

	When :code:`problem_type=None`, the nature of the supervised learning problem (i.e. regression or classification) is inferred from whether or not :code:`prediction_column` is categorical.


	Parameters
	----------
	data_df : pandas.DataFrame
		The pandas DataFrame containing the data.
	prediction_column : str
		The name of the column containing true labels.
	problem_type : None | 'classification' | 'regression'
		The type of supervised learning problem. When None, it is inferred from the column type and the number of distinct values.



	Returns
	-------
	result : pandas.DataFrame
		The result is a pandas.Dataframe with columns (where applicable):

		* :code:`'Selection Order'`: The order in which the associated variable was selected, starting at 1 for the most important variable.
		* :code:`'Variable'`: The column name corresponding to the input variable.
		* :code:`'Running Achievable R-Squared'`: The highest :math:`R^2` that can be achieved by a classification model using all variables selected so far, including this one.
		* :code:`'Running Achievable Accuracy'`: The highest classification accuracy that can be achieved by a classification model using all variables selected so far, including this one.
		* :code:`'Running Achievable RMSE'`: The highest classification accuracy that can be achieved by a classification model using all variables selected so far, including this one.


	.. admonition:: Theoretical Foundation

		Section :ref:`a) Model Explanation`.

	"""
	assert prediction_column in data_df.columns, 'The label column should be a column of the dataframe.'
	assert problem_type.lower() in ['classification', 'regression']
	if problem_type.lower() == 'regression':
		assert np.can_cast(data_df[prediction_column], float), 'The prediction column should be numeric'

	k = 0
	kp = 0
	max_k = 100

	file_name = upload_data(data_df)
	spinner = Halo(text='Waiting for results from the backend.', spinner='dots')
	spinner.start()

	if file_name:
		job_id = EXPLANATION_JOB_IDS.get((file_name, prediction_column, problem_type), None)
		if job_id:
			api_response = APIClient.route(
				path='/wk/variable-selection', method='POST', \
				file_name=file_name, target_column=prediction_column, \
				problem_type=problem_type, timestamp=int(time()), job_id=job_id, \
				snr=snr)
		else:
			api_response = APIClient.route(
				path='/wk/variable-selection', method='POST', \
				file_name=file_name, target_column=prediction_column, \
				problem_type=problem_type, timestamp=int(time()), snr=snr)

		initial_time = time()
		while api_response.status_code == requests.codes.ok and k < max_k:
			if kp%2 != 0:
				sleep(2 if kp<5 else 10 if k < max_k-4 else 300)
				kp += 1
				k = kp//2

			else:
				try:
					response = api_response.json()
					if 'job_id' in response:
						job_id = response['job_id']
						EXPLANATION_JOB_IDS[(file_name, prediction_column, problem_type)] = job_id
						sleep(2 if kp<5 else 10 if k < max_k-4 else 300)
						kp += 1
						k = kp//2

						# Note: it is important to pass the job_id to avoid being charged twice for the work.
						api_response = APIClient.route(
							path='/wk/variable-selection', method='POST', \
							file_name=file_name, target_column=prediction_column, \
							problem_type=problem_type, timestamp=int(time()), job_id=job_id, \
							snr=snr)

						try:
							response = api_response.json()
							if 'eta' in response:
								progress_text = '%s%% Completed.' % response['progress_pct'] if 'progress_pct' in response else ''
								spinner.text = 'Waiting for results from the backend. ETA: %s. %s' % (response['eta'], progress_text)
						except:
							pass

					if 'job_id' not in response:
						duration = int(time()-initial_time)
						duration = str(duration) + 's' if duration < 60 else str(duration//60) + 'min'

						result = {}

						if 'selection_order' in response:
							result['Selection Order'] = response['selection_order']				

						if 'variable' in response:
							result['Variable'] = response['variable']	

						if 'r-squared' in response:
							result['Running Achievable R-Squared'] = response['r-squared']

						if 'log-likelihood' in response:
							result['Running Achievable Log-Likelihood Per Sample'] = response['log-likelihood']

						if 'rmse' in response and problem_type.lower() == 'regression':
							result['Running Achievable RMSE'] = response['rmse']

						if 'accuracy' in response and problem_type.lower() == 'classification':
							result['Running Achievable Accuracy'] = response['accuracy']

						result = pd.DataFrame.from_dict(result)

						if 'selection_order' in response:
							result.set_index('Selection Order', inplace=True)

						spinner.text = 'Received results from the backend after %s.' % duration
						spinner.succeed()
						return result


				except:
					logging.exception('\nModel explanation failed. Last HTTP code: %s, Content: %s' % (api_response.status_code, api_response.content))
					spinner.text = 'The backend encountered an unexpected error we are looking into. Please try again later.'
					spinner.fail()
					return None

		if api_response.status_code != requests.codes.ok:
			spinner.text = 'The backend is taking longer than expected. Please try again later'
			spinner.fail()
			try:
				response = api_response.json()
				if 'message' in response:
					logging.error('\n%s' % response['message'])
			except:
				logging.error('\nModel explanation failed. Last HTTP code: %s, Content: %s' % (api_response.status_code, api_response.content))

	return None





#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Estimation of the amount by which the performance of a trained supervised learning model can be increased, either in a model-driven fashion, or a data-driven fashion.
"""
import json
import logging
logging.basicConfig(level=logging.INFO)
import requests
import sys
from time import time, sleep

import numpy as np
import pandas as pd

try:
	get_ipython().__class__.__name__
	from halo import HaloNotebook as Halo
except:
	from halo import Halo

from kxy.api import APIClient, upload_data

# Cache old job ids to avoid being charged twice for the same job.
DD_IMPROVABILITY_JOB_IDS = {}
MD_IMPROVABILITY_JOB_IDS = {}

def data_driven_improvability(data_df, target_column, new_variables, problem_type, snr='auto'):
	"""
	.. data-driven-improvability:
	Estimate the potential performance boost that a set of new explanatory variables can bring about.


	Parameters
	----------
	data_df : pandas.DataFrame
		The pandas DataFrame containing the data.
	target_column : str
		The name of the column containing true labels.
	new_variables : list
		The names of the columns to use as new explanatory variables.
	problem_type : None | 'classification' | 'regression'
		The type of supervised learning problem. When None, it is inferred from whether or not :code:`target_column` is categorical.



	Returns
	-------
	result : pandas.Dataframe
		The result is a pandas.Dataframe with columns (where applicable):

		* :code:`'Accuracy Boost'`: The classification accuracy boost that the new explanatory variables can bring about.
		* :code:`'R-Squared Boost'`: The :math:`R^2` boost that the new explanatory variables can bring about.
		* :code:`'RMSE Reduction'`: The reduction in Root Mean Square Error that the new explanatory variables can bring about.
		* :code:`'Log-Likelihood Per Sample Boost'`: The boost in log-likelihood per sample that the new explanatory variables can bring about.


	.. admonition:: Theoretical Foundation

		Section :ref:`3 - Model Improvability`.
		
	"""
	assert target_column in data_df.columns, 'The label column should be a column of the dataframe.'
	assert problem_type.lower() in ['classification', 'regression']
	assert len(new_variables) > 0, 'New variables should be provided'
	for col in new_variables:
		assert col in data_df.columns, '%s should be a column in the dataframe' % col
	if problem_type.lower() == 'regression':
		assert np.can_cast(data_df[target_column], float), 'The target column should be numeric'

	k = 0
	kp = 0
	max_k = 100

	file_name = upload_data(data_df)
	spinner = Halo(text='Waiting for results from the backend.', spinner='dots')
	spinner.start()

	if file_name:
		job_id = DD_IMPROVABILITY_JOB_IDS.get((file_name, target_column, str(new_variables), problem_type, snr), None)

		if job_id:
			api_response = APIClient.route(
				path='/wk/data-driven-improvability', method='POST', \
				file_name=file_name, target_column=target_column, \
				problem_type=problem_type, new_variables=json.dumps(new_variables), \
				job_id=job_id, timestamp=int(time()), snr=snr)
		else:
			api_response = APIClient.route(
				path='/wk/data-driven-improvability', method='POST', \
				file_name=file_name, target_column=target_column, \
				problem_type=problem_type, new_variables=json.dumps(new_variables), \
				timestamp=int(time()), snr=snr)


		initial_time = time()
		while api_response.status_code == requests.codes.ok and k < max_k:
			if kp%2 != 0:
				sleep(2 if kp<5 else 10 if k < max_k-4 else 300)
				kp += 1
				k = kp//2

			else:
				try:
					response = api_response.json()
					if 'job_id' in response:
						job_id = response['job_id']
						DD_IMPROVABILITY_JOB_IDS[(file_name, target_column, str(new_variables), problem_type, snr)] = job_id
						sleep(2 if kp<5 else 10 if k < max_k-4 else 300)
						kp += 1
						k = kp//2
						api_response = APIClient.route(
							path='/wk/data-driven-improvability', method='POST', \
							file_name=file_name, target_column=target_column, \
							problem_type=problem_type, new_variables=json.dumps(new_variables), \
							timestamp=int(time()), snr=snr)

						try:
							response = api_response.json()
							if 'eta' in response:
								progress_text = '%s%% Completed.' % response['progress_pct'] if 'progress_pct' in response else ''
								spinner.text = 'Waiting for results from the backend. ETA: %s. %s' % (response['eta'], progress_text)
						except:
							pass

					if 'job_id' not in response:
						duration = int(time()-initial_time)
						duration = str(duration) + 's' if duration < 60 else str(duration//60) + 'min'
						result = {}
						if 'r-squared-boost' in response:
							result['R-Squared Boost'] = [response['r-squared-boost']]

						if 'log-likelihood-boost' in response:
							result['Log-Likelihood Per Sample Boost'] = [response['log-likelihood-boost']]

						if 'rmse-reduction' in response and problem_type.lower() == 'regression':
							result['RMSE Reduction'] = [response['rmse-reduction']]

						if 'accuracy-boost' in response and problem_type.lower() == 'classification':
							result['Accuracy Boost'] = [response['accuracy-boost']]

						result = pd.DataFrame.from_dict(result)
						spinner.text = 'Received results from the backend after %s' % duration
						spinner.succeed()
						return result

				except:
					spinner.text = 'The backend encountered an unexpected error we are looking into. Please try again later.'
					spinner.fail()
					return None

		if api_response.status_code != requests.codes.ok:
			spinner.text = 'The backend is taking longer than expected. Try again later.'
			spinner.fail()
			try:
				response = api_response.json()
				if 'message' in response:
					logging.error('\n%s' % response['message'])
			except:
				logging.error('\nData-driven improvability failed. Last HTTP code: %s' % api_response.status_code)

	return None


def model_driven_improvability(data_df, target_column, prediction_column, problem_type, snr='auto'):
	"""
	.. model-driven-improvability:
	Estimate the extent to which a trained supervised learner may be improved in a model-driven fashion (i.e. without resorting to additional explanatory variables).


	Parameters
	----------
	data_df : pandas.DataFrame
		The pandas DataFrame containing the data.
	target_column : str
		The name of the column containing true labels.
	prediction_column : str
		The name of the column containing model predictions.
	problem_type : None | 'classification' | 'regression'
		The type of supervised learning problem. When None, it is inferred from whether or not :code:`target_column` is categorical.



	Returns
	-------
	result : pandas.Dataframe
		The result is a pandas.Dataframe with columns (where applicable):

		* :code:`'Lost Accuracy'`: The amount of classification accuracy that was irreversibly lost when training the supervised learner.
		* :code:`'Lost R-Squared'`: The amount of :math:`R^2` that was irreversibly lost when training the supervised learner.
		* :code:`'Lost RMSE'`: The amount of Root Mean Square Error that was irreversibly lost when training the supervised learner.		
		* :code:`'Lost Log-Likelihood Per Sample'`: The amount of true log-likelihood per sample that was irreversibly lost when training the supervised learner.

		* :code:`'Residual R-Squared'`: For regression problems, this is the highest :math:`R^2` that may be achieved when using explanatory variables to predict regression residuals.
		* :code:`'Residual RMSE'`: For regression problems, this is the lowest Root Mean Square Error that may be achieved when using explanatory variables to predict regression residuals.
		* :code:`'Residual Log-Likelihood Per Sample'`: For regression problems, this is the highest log-likelihood per sample that may be achieved when using explanatory variables to predict regression residuals.


	.. admonition:: Theoretical Foundation

		Section :ref:`3 - Model Improvability`.

	"""
	assert target_column in data_df.columns, 'The label column should be a column of the dataframe.'
	assert prediction_column in data_df.columns, 'The prediction column should be a column of the dataframe.'
	assert problem_type.lower() in ['classification', 'regression']
	if problem_type.lower() == 'regression':
		assert np.can_cast(data_df[target_column], float), 'The target column should be numeric'
		assert np.can_cast(data_df[prediction_column], float), 'The prediction column should be numeric'

	k = 0
	kp = 0
	max_k = 100

	file_name = upload_data(data_df)
	spinner = Halo(text='Waiting for results from the backend.', spinner='dots')
	spinner.start()

	if file_name:
		job_id = MD_IMPROVABILITY_JOB_IDS.get((file_name, target_column, prediction_column, problem_type, snr), None)

		if job_id:
			api_response = APIClient.route(
				path='/wk/model-driven-improvability', method='POST', \
				file_name=file_name, target_column=target_column, \
				problem_type=problem_type, prediction_column=prediction_column, \
				job_id=job_id, timestamp=int(time()), snr=snr)
		else:
			api_response = APIClient.route(
				path='/wk/model-driven-improvability', method='POST', \
				file_name=file_name, target_column=target_column, \
				problem_type=problem_type, prediction_column=prediction_column, \
				timestamp=int(time()), snr=snr)

		while api_response.status_code == requests.codes.ok and k <= max_k:
			if kp%2 != 0:
				sleep(2 if kp<5 else 10 if k < max_k-4 else 300)
				kp += 1
				k = kp//2

			else:
				try:
					response = api_response.json()
					if 'job_id' in response:
						job_id = response['job_id']
						MD_IMPROVABILITY_JOB_IDS[(file_name, target_column, prediction_column, problem_type, snr)] = job_id
						sleep(2 if kp<5 else 10 if k < max_k-4 else 300)
						kp += 1
						k = kp//2
						api_response = APIClient.route(
							path='/wk/model-driven-improvability', method='POST', \
							file_name=file_name, target_column=target_column, \
							problem_type=problem_type, prediction_column=prediction_column, \
							job_id=job_id, timestamp=int(time()), snr=snr)

						try:
							response = api_response.json()
							if 'eta' in response:
								progress_text = '%s%% Completed.' % response['progress_pct'] if 'progress_pct' in response else ''
								spinner.text = 'Waiting for results from the backend. ETA: %s. %s' % (response['eta'], progress_text)
						except:
							pass

					if 'job_id' not in response:
						result = {}

						if 'lost-r-squared' in response:
							result['Lost R-Squared'] = [response['lost-r-squared']]			

						if 'lost-log-likelihood' in response:
							result['Lost Log-Likelihood Per Sample'] = [response['lost-log-likelihood']]

						if 'lost-rmse' in response and problem_type.lower() == 'regression':
							result['Lost RMSE'] = [response['lost-rmse']]

						if 'lost-accuracy' in response and problem_type.lower() == 'classification':
							result['Lost Accuracy'] = [response['lost-accuracy']]


						if problem_type.lower() == 'regression':
							if 'residual-r-squared' in response:
								result['Residual R-Squared'] = [response['residual-r-squared']]			

							if 'residual-log-likelihood' in response:
								result['Residual Log-Likelihood Per Sample'] = [response['residual-log-likelihood']]

							if 'residual-rmse' in response:
								result['Residual RMSE'] = [response['residual-rmse']]

						result = pd.DataFrame.from_dict(result)
						spinner.text = 'Received results from the backend after %s' % duration
						spinner.succeed()
						return result

				except:
					spinner.text = 'The backend encountered an unexpected error we are looking into. Please try again later.'
					spinner.fail()
					return None

		if api_response.status_code != requests.codes.ok:
			spinner.text = 'The backend is taking longer than expected. Please try again later.'
			spinner.fail()
			try:
				response = api_response.json()
				if 'message' in response:
					logging.error('\n%s' % response['message'])
			except:
				logging.error('\nModel-driven improvability failed. Last HTTP code: %s' % api_response.status_code)

	return None




